package edu.stanford.cs247.util;

import java.io.IOException;

import android.content.Context;
import android.hardware.Camera;
import android.hardware.Camera.PreviewCallback;
import android.view.SurfaceHolder;
import android.view.SurfaceView;

/**
 * A class which reads video frames from the camera and invokes a callback
 * when each frame is available.
 * 
 * <p>To use this class, your app must have permission CAMERA.</p>
 */
public class VideoReader {

    Camera mCamera;
    boolean mFinished;
    CameraData mData;
    Listener mListener;
    Context mContext;
    SurfaceView mSurface;
    
    public VideoReader(Context context, Listener listener) {
    	mSurface = new VideoSurface(context);
        mData = new CameraData();
        mListener = listener;
        mFinished = false;
    }
    
    public SurfaceView surface() {
    	return mSurface;
    }
    
    /**
     * Starts reading camera frames.
     */
    public void startReader() {
    	if (mCamera != null) return;
    	mFinished = false;
    	
        mCamera = Camera.open();
    	
        // Preview callback used whenever a new frame is available
        mCamera.setPreviewCallback(new PreviewCallback() {
        	public void onPreviewFrame(byte[] data, Camera camera) {
        		if (mData == null || mFinished) return;
    		  
        		if (mData.mYUVData == null) {
        			// Initialize the camera data struct
        			Camera.Parameters params = camera.getParameters();
        			mData.mWidth = params.getPreviewSize().width;
        			mData.mHeight = params.getPreviewSize().height;
        			mData.mRGBData = new int[mData.mWidth * mData.mHeight]; 
        			mData.mYUVData = new byte[data.length];        			  
        		}
    		  
        		// Pass YUV data to draw-on-top companion
        		System.arraycopy(data, 0, mData.mYUVData, 0, data.length);

        		// Convert from YUV to RGB
        		decodeYUV420SP(mData.mRGBData, mData.mYUVData,
        				mData.mWidth, mData.mHeight);
    		  
        		mListener.onFrame(mData);
        	}
        });
        
        update();
    }
    
    protected void update() {
        Camera.Parameters parameters = mCamera.getParameters();
        parameters.setPreviewSize(320, 240);
        parameters.setPreviewFrameRate(15);
        parameters.setSceneMode(Camera.Parameters.SCENE_MODE_NIGHT);
        parameters.setFocusMode(Camera.Parameters.FOCUS_MODE_AUTO);
        mCamera.setParameters(parameters);
        mCamera.startPreview();
    }

    /**
     * Stops reading camera frames.
     */
    public void stopReader() {
    	if (mCamera == null) return;
    	mFinished = true;
    	mCamera.setPreviewCallback(null);
        mCamera.stopPreview();
        mCamera.release();
        mCamera = null;
    }
        
    /**
     * Converts a YUV format image to RGB pixels.
     */
    public static void decodeYUV420SP(int[] rgb, byte[] yuv420sp, int width, int height) {
    	final int frameSize = width * height;
    	
    	for (int j = 0, yp = 0; j < height; j++) {
    		int uvp = frameSize + (j >> 1) * width, u = 0, v = 0;
    		for (int i = 0; i < width; i++, yp++) {
    			int y = (0xff & ((int) yuv420sp[yp])) - 16;
    			if (y < 0) y = 0;
    			if ((i & 1) == 0) {
    				v = (0xff & yuv420sp[uvp++]) - 128;
    				u = (0xff & yuv420sp[uvp++]) - 128;
    			}
    			
    			int y1192 = 1192 * y;
    			int r = (y1192 + 1634 * v);
    			int g = (y1192 - 833 * v - 400 * u);
    			int b = (y1192 + 2066 * u);
    			
    			if (r < 0) r = 0; else if (r > 262143) r = 262143;
    			if (g < 0) g = 0; else if (g > 262143) g = 262143;
    			if (b < 0) b = 0; else if (b > 262143) b = 262143;
    			
    			rgb[yp] = 0xff000000 | ((r << 6) & 0xff0000) | ((g >> 2) & 0xff00) | ((b >> 10) & 0xff);
    		}
    	}
    }
    
    /**
     * A struct containing data for each camera frame.
     */
    public static class CameraData {
    	public byte[] mYUVData;
    	public int[] mRGBData;
    	public int mWidth;
    	public int mHeight;
    }
    
    /**
     * An interface for listening for frame events.
     */
    public interface Listener {
    	public void onFrame(CameraData data);
    }
    
    public class VideoSurface extends SurfaceView
    	implements SurfaceHolder.Callback
    {
    	SurfaceHolder mHolder;
    	
    	public VideoSurface(Context context) {
            super(context);
    	    
    		// Install a SurfaceHolder.Callback so we get notified when the
            // underlying surface is created and destroyed.
            mHolder = getHolder();
            mHolder.addCallback(this);
            mHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);
    	}
    	
	    public void surfaceCreated(SurfaceHolder holder) {
	    	if (mCamera == null) startReader();
	    	
	    	try {
	    		mCamera.setPreviewDisplay(holder);
	    	} catch (IOException exception) {
	            mCamera.release();
	            mCamera = null;
	        }
	    	
	    }
	
	    public void surfaceDestroyed(SurfaceHolder holder) {
	        // Surface will be destroyed when we return, so stop the preview.
	        // Because the CameraDevice object is not a shared resource, it's very
	        // important to release it when the activity is paused.
	    	stopReader();
	    }
	
	    public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
	        if (mCamera != null) update();
	    }
    }

}